PROGRAM Clique (Infile, OUTPUT);
   (* version 2.6 *)

CONST NmLngth = 10; (* number of characters in a name *)
   MaxSpp = 50;     (* max. number of species *)
   MaxChr = 50;     (* max. number of characters *)
   FormWide = 80;   (* width of output page *)

TYPE Vector = ARRAY [1..MaxChr] OF BOOLEAN;
   NameArray = ARRAY [1..MaxSpp, 1..NmLngth] OF CHAR;
   DataArray = ARRAY [1..MaxSpp, 1..MaxChr] OF BOOLEAN;
   Matrix = ARRAY [1..MaxChr] OF Vector;

VAR NumSpp, NumChars, Cliqmin : INTEGER;
   Clmin : BOOLEAN;
   Name : NameArray;
   Data : DataArray;
   Comp : Matrix; (* the character compatibility matrix *)
   Infile : TEXT;

PROCEDURE NewLine (i, j : INTEGER);
   (* goes to new line if i MOD j is zero *)
   VAR k : INTEGER;
BEGIN (* NewLine *)
   IF ((i MOD j) = 0)
   THEN BEGIN
      WRITELN(OUTPUT);
      FOR k := 1 TO NmLngth+2 DO
         WRITE(OUTPUT, ' ');
      END;
END; (* NewLine *)

PROCEDURE ReadData (VAR NumSpp, NumChars : INTEGER; VAR
          Name : NameArray; VAR Data : DataArray);
   (* reads the species names and character data *)
   VAR i, j : INTEGER;
      ch : CHAR;
      badchrs : BOOLEAN;
BEGIN (* ReadData *)
   badchrs := FALSE;
   READ (Infile, NumSpp, NumChars);
   Clmin := FALSE;
   WHILE NOT EOLN(Infile) DO BEGIN
      READ (Infile, ch);
      IF (ch = 'C') OR (ch = 'c')
      THEN Clmin := TRUE;
      END;
   READLN (Infile);
   IF Clmin
   THEN BEGIN
      READ (Infile, ch);
      READLN (Infile, Cliqmin);
      END;
   WRITELN(OUTPUT);
   WRITE(OUTPUT, ' Largest Clique program, ');
   WRITELN(OUTPUT, ' version 2.6');
   WRITELN(OUTPUT);
   WRITELN(OUTPUT, ' ', NumSpp:3,' species,',
               NumChars:3,' character states');
   WRITELN(OUTPUT);
   WRITELN(OUTPUT, ' ', 'Species     Character states');
   WRITELN(OUTPUT, ' ', '-------     ----------------');
   FOR i := 1 TO NumSpp DO BEGIN
      WRITE(OUTPUT, ' ');
      FOR j := 1 TO NmLngth DO BEGIN
         READ (Infile, ch); WRITE(OUTPUT, ch);
         Name [i, j] := ch;
         END;
      WRITE(OUTPUT, '  ');
      FOR j := 1 TO NumChars DO BEGIN
         IF EOLN (Infile) THEN READLN (Infile);
         READ (Infile, ch); WRITE(OUTPUT, ch);
         NewLine(j, 50);
         badchrs := badchrs OR ( NOT ((ch = '0') OR (ch = '1')) );
         IF (j MOD 5 = 0)
         THEN WRITE(OUTPUT, ' ');
         Data [i, j] := (ch = '1');
         END;
      READLN (Infile); WRITELN(OUTPUT);
      END;
   WRITELN(OUTPUT);
   IF badchrs
   THEN BEGIN
      WRITE(OUTPUT, ' WARNING -- STATES OTHER THAN');
      WRITELN(OUTPUT, ' 0 OR 1 PRESENT IN DATA');
      WRITELN(OUTPUT);
      END;
END; (* ReadData *)

PROCEDURE SetUp (VAR Comp : Matrix);
   (* sets up the compatibility matrix *)
   VAR i, j : INTEGER;

   FUNCTION Compatible (ch1, ch2 : INTEGER) : BOOLEAN;
      (* TRUE if two characters are compatible *)
      VAR i : INTEGER;
         Info : ARRAY [1..4] OF BOOLEAN;
   BEGIN (* Compatible *)
      FOR i := 1 TO 4 DO Info [i] := FALSE;
      FOR i := 1 TO NumSpp DO
         IF Data[i, ch1] AND Data[i, ch2]
         THEN Info [1] := TRUE
         ELSE IF Data [i, ch1] AND NOT Data [i, ch2]
            THEN Info [2] := TRUE
            ELSE IF NOT Data [i, ch1] AND Data [i, ch2]
               THEN Info [3] := TRUE
               ELSE Info [4] := TRUE;
      Compatible := NOT (Info [1] AND Info [2] AND
                         Info[3] AND Info [4]);
   END; (* Compatible *)

BEGIN (* SetUp *)
   WRITELN(OUTPUT, ' Character Compatibility Matrix');
   WRITELN(OUTPUT, ' --------- ------------- ------');
   FOR i:= 1 TO NumChars DO BEGIN
      WRITE(OUTPUT, ' ');
      FOR j := 1 TO i-1 DO BEGIN
         IF Comp [i][j]
         THEN WRITE(OUTPUT, '1')
         ELSE WRITE(OUTPUT, '0');
         NewLine (j, 70);
         END;
      Comp [i][i] := TRUE;
      WRITE(OUTPUT, '1');
      FOR j := i+1 TO NumChars DO BEGIN
         Comp [i][j] := Compatible (i, j);
         IF Comp [i][j]
         THEN WRITE(OUTPUT, '1')
         ELSE WRITE(OUTPUT, '0');
         Comp [j][i] := Comp [i][j];
         END;
      WRITELN(OUTPUT);
      END;
   WRITELN(OUTPUT);
END; (* Setup *)

PROCEDURE DoAll (Chars : Vector; VAR Processed, Rarer : Vector);
   TYPE SpVector = ARRAY [1..MaxSpp] OF INTEGER;
      ChVector = ARRAY [1..MaxChr] OF INTEGER;
   VAR i, j, MaxChars : INTEGER;
      SpOrder : SpVector;
      ChOrder, Count : ChVector;

   FUNCTION RarerState (i : INTEGER) : BOOLEAN;
      (* True if the rarer state for the i-th
         character is 1 *)
      VAR j, ZeroCount, OneCount : INTEGER;
   BEGIN (* RarerState *)
      ZeroCount :=0;
      OneCount := 0;
      FOR j := 1 TO NumSpp DO
         IF Data [j, i]
         THEN OneCount := OneCount + 1
         ELSE ZeroCount := ZeroCount + 1;
      RarerState := (OneCount < ZeroCount);
   END; (* RarerState *)

   PROCEDURE SpSort (VAR SpOrder : SpVector; i : INTEGER);
      (* sorts the species according to the rarer
         state of the i-th character *)
      VAR j, ptr1, ptr2 : INTEGER;
         V1, V2 : SpVector;
   BEGIN (* SpSort *)
      ptr1 := 0;
      ptr2 := 0;
      FOR j := 1 TO NumSpp DO
         IF Data [SpOrder [j], i]
         THEN BEGIN
            ptr1 := ptr1 + 1;
            V1 [ptr1] := SpOrder [j];
            END
         ELSE BEGIN
            ptr2 := ptr2 + 1;
            V2 [ptr2] := SpOrder [j];
            END;
      FOR j := 1 TO ptr1 DO SpOrder [j] := V1 [j];
      FOR j := 1 TO ptr2 DO SpOrder [j+ptr1] := V2 [j];
   END; (* SpSort *)

   PROCEDURE Init (VAR ChOrder, Count : ChVector; VAR
              MaxChars : INTEGER);
      (* initialize vectors and character count *)
      VAR i, j, temp : INTEGER;
   BEGIN (* Init *)
      MaxChars := 0;
      FOR i := 1 TO NumChars DO
         IF Chars [i]
         THEN BEGIN
            MaxChars := MaxChars + 1;
            ChOrder [MaxChars] := i;
            temp := 0;
            FOR j := 1 TO NumSpp DO
               IF Data [j, i] THEN temp := temp + 1;
            IF temp > NumSpp - temp
            THEN temp := NumSpp - temp;
            Count [i] := temp;
            END;
   END; (* Init *)

   PROCEDURE ChSort (VAR ChOrder : ChVector; Count : ChVector);
      (* sorts the characters by cardinality *)
      VAR j, temp : INTEGER;
         ordered : BOOLEAN;
   BEGIN (* ChSort *)
      ordered := FALSE;
      WHILE NOT ordered DO BEGIN
         ordered := TRUE;
         FOR j := 1 TO MaxChars-1 DO
         IF Count[ChOrder[j]] < Count[ChOrder[j+1]]
         THEN BEGIN
            ordered := FALSE;
            temp := ChOrder [j];
            ChOrder [j] := ChOrder [j+1];
            ChOrder [j+1] := temp;
            END;
         END;
   END; (* ChSort *)

   PROCEDURE PrintClique (Chars : Vector);
      (* prints the characters in a clique *)
      VAR i, j : INTEGER;
   BEGIN (* PrintClique *)
      WRITELN(OUTPUT);
      WRITE(OUTPUT, ' Characters: (');
      j := 0;
      FOR i := 1 TO NumChars DO BEGIN
         IF Chars [i]
         THEN BEGIN
            WRITE(OUTPUT, i:3);
            j := j + 1;
            NewLine (j , (FormWide-15) DIV 3);
            END;
         END;
      WRITELN(OUTPUT, ')');
      WRITELN(OUTPUT);
   END; (* PrintClique *)

   PROCEDURE TreeTable (SpOrder : SpVector; ChOrder : ChVector);
      VAR i, j, k, l, n, temp : INTEGER;
         ch : CHAR;
         Spaces : ARRAY [0..MaxChr] OF INTEGER;

      FUNCTION Next (R, C : INTEGER) : INTEGER;
      BEGIN (* Next *)
         IF C = MaxChars
         THEN Next := MaxChars + 1
         ELSE BEGIN
            REPEAT
               IF C < MaxChars
               THEN C := C + 1
            UNTIL (C = MaxChars) OR
                      (Data [SpOrder[R], ChOrder[C]]);
            IF C = MaxChars
            THEN IF NOT Data[SpOrder[R], ChOrder[C]]
               THEN C := C + 1;
            Next := C;
            END;
      END; (* Next *)

   BEGIN (* TreeTable *)
      n := 3;
      IF (MaxChars*3 + 13) > FormWide
      THEN n := (FormWide-13) DIV MaxChars;
      WRITE(OUTPUT, '  ');
      FOR i := 1 TO MaxChars DO BEGIN
         WRITE(OUTPUT, ChOrder [i]:3);
         NewLine (i, (FormWide-12) DIV 3);
         END;
      WRITELN(OUTPUT);
      WRITELN(OUTPUT);
      FOR i := 0 TO MaxChars DO
         Spaces [i] := i;
      FOR i := 1 TO NumSpp DO BEGIN
         WRITE(OUTPUT, ' !');
         j := 0;
         REPEAT
            temp := Next (i, j);
            IF (temp <> Spaces [j]) OR
                (temp = MaxChars + 1)
            THEN ch := '-'
            ELSE ch := ' ';
            FOR k := j+1 TO temp-1 DO BEGIN
               FOR l := 1 TO n DO
                  WRITE(OUTPUT, ch);
               NewLine (k, FormWide DIV n);
               END;
            IF temp <> MaxChars+1
            THEN BEGIN
               FOR l := 1 TO n-1 DO
                  WRITE(OUTPUT, ch);
               IF Rarer [ChOrder [temp]]
               THEN WRITE(OUTPUT, '1')
               ELSE WRITE(OUTPUT, '0');
               END;
            Spaces [j] := temp;
            j := temp;
            NewLine (j, FormWide DIV n);
         UNTIL j = MaxChars + 1;
         WRITE(OUTPUT, ' ');
         FOR j := 1 TO NmLngth DO
            WRITE(OUTPUT, Name [SpOrder [i], j]);
         WRITELN(OUTPUT);
         END;
      WRITELN(OUTPUT);
   END; (* Treeable *)

BEGIN (* DoAll *)
   Init (ChOrder, Count, MaxChars);
   ChSort (ChOrder, Count);
   FOR i := 1 TO NumSpp DO SpOrder [i] := i;
   FOR i := 1 TO NumChars DO
      IF Chars [i]
      THEN BEGIN
        IF NOT Processed [i]
        THEN BEGIN
           Rarer [i] := RarerState (i);
           FOR j := 1 TO NumSpp DO
              Data [j, i] := (Data [j, i] = Rarer [i]);
           Processed [i] := TRUE;
           END;
        END;
   FOR i := MaxChars DOWNTO 1 DO
      SpSort (SpOrder, ChOrder[i]);
   PrintClique (Chars);
   TreeTable (SpOrder, ChOrder);
END; (* DoAll *)

PROCEDURE GetMaxCliques (VAR Comp : Matrix);
   (* recursively generates the largest cliques *)
   VAR i : INTEGER;
       Chars, Temp, Processed, Rarer,
                Candidates, Excluded : Vector;

   PROCEDURE Intersect (V1, V2 : Vector; VAR V3 : Vector);
      (* takes the logical intersection V1 AND V2 *)
      VAR i : INTEGER;
   BEGIN (* Intersect *)
      FOR i := 1 TO NumChars DO
         V3 [i] := V1 [i] AND V2 [i];
   END; (* Intersect *)

   FUNCTION Count (V : Vector) : INTEGER;
      (* counts the 1's in V *)
      VAR i, TempCount : INTEGER;
   BEGIN (* Count *)
      TempCount := 0;
      FOR i := 1 TO NumChars DO
         IF V [i] THEN TempCount := TempCount + 1;
      Count := TempCount;
   END; (* Count *)

   PROCEDURE Gen1 (i, CurSize : INTEGER; Chars, Candidates,
                    Excluded : Vector);
      (* finds largest size cliques and prints them out *)
      VAR j, k, Actual, Possible : INTEGER;
         Futile : BOOLEAN;
         Cprime, Exprime : Vector;
   BEGIN (* Gen1 *)
      j := i;
      WHILE (j <= NumChars) DO BEGIN
         IF Candidates[j]
         THEN BEGIN
            Chars[j] := TRUE;
            Candidates[j] := FALSE;
            CurSize := CurSize + 1;
            Possible := Count (Candidates);
            Intersect (Candidates, Comp[j], Cprime);
            Actual := Count (Cprime);
            Intersect (Excluded, Comp[j], Exprime);
            Futile := FALSE;
            FOR k := 1 TO j-1 DO BEGIN
               IF (Exprime[k] AND (NOT Futile))
               THEN BEGIN
                  Intersect (Cprime, Comp[k], Temp);
                  Futile := (Count(Temp) = Actual);
                  END;
               END;
            IF (CurSize + Actual >= Cliqmin) AND (NOT Futile)
            THEN IF Actual > 0
               THEN Gen1 (j+1, CurSize, Chars, Cprime, Exprime)
               ELSE IF CurSize > Cliqmin
                  THEN Cliqmin := CurSize;
            IF (Possible > Actual)
            THEN BEGIN
               Chars[j] := FALSE;
               Excluded[j] := TRUE;
               CurSize := CurSize - 1;
               END
            ELSE j := NumChars;
            END;
         j := j + 1;
         END;
   END; (* Gen1 *)

   PROCEDURE Gen2 (i, CurSize : INTEGER; Chars, Candidates,
                    Excluded : Vector);
      (* finds largest size cliques and prints them out *)
      VAR j, k, Actual, Possible : INTEGER;
         Futile : BOOLEAN;
         Cprime, Exprime : Vector;
   BEGIN (* Gen2 *)
      j := i;
      WHILE (j <= NumChars) DO BEGIN
         IF Candidates[j]
         THEN BEGIN
            Chars[j] := TRUE;
            Candidates[j] := FALSE;
            CurSize := CurSize + 1;
            Possible := Count (Candidates);
            Intersect (Candidates, Comp[j], Cprime);
            Actual := Count (Cprime);
            Intersect (Excluded, Comp[j], Exprime);
            Futile := FALSE;
            FOR k := 1 TO j-1 DO BEGIN
               IF (Exprime[k] AND (NOT Futile))
               THEN BEGIN
                  Intersect (Cprime, Comp[k], Temp);
                  Futile := (Count(Temp) = Actual);
                  END;
               END;
            IF (CurSize + Actual >= Cliqmin) AND (NOT Futile)
            THEN IF Actual > 0
               THEN Gen2 (j+1, CurSize, Chars, Cprime, Exprime)
               ELSE DoAll (Chars, Processed, Rarer);
            IF (Possible > Actual)
            THEN BEGIN
               Chars[j] := FALSE;
               Excluded[j] := TRUE;
               CurSize := CurSize - 1;
               END
            ELSE j := NumChars;
            END;
         j := j + 1;
         END;
   END; (* Gen2 *)

   BEGIN (* GetMaxCliques *)
      IF Clmin
      THEN BEGIN
         WRITELN(OUTPUT, ' Cliques with at least',
            Cliqmin:3, ' characters');
         WRITELN(OUTPUT, ' ------- ---- -- -----',
                    ' -- ----------');
         END
      ELSE BEGIN
         Cliqmin := 0;
         WRITELN(OUTPUT, ' Largest Cliques');
         WRITELN(OUTPUT, ' ------- -------');
         FOR i := 1 TO NumChars DO BEGIN
            Chars [i] := FALSE;
            Excluded [i] := FALSE;
            Candidates [i] := TRUE;
            END;
         Gen1 (1, 0, Chars, Candidates, Excluded);
         END;
      FOR i := 1 TO NumChars DO BEGIN
         Chars [i] := FALSE;
         Candidates [i] := TRUE;
         Processed [i] := FALSE;
         Excluded [i] := FALSE;
         END;
      Gen2 (1, 0, Chars, Candidates, Excluded);
      WRITELN(OUTPUT);
   END; (* GetMaxCliques *)

BEGIN (* Main Program *)
   RESET (Infile);
   ReadData (NumSpp, NumChars, Name, Data);
   SetUp (Comp);
   GetMaxCliques (Comp);
END.

