PROGRAM Dnaml(infile, OUTPUT);
   (* version 2.61 *)

   CONST maxsp2 = 38;    (* 2 x maximum number of species - 2 *)
      maxsites = 120;    (* maximum number of sites *)
      smoothings = 4;    (* number of passes through smoothing algorithm *)
      iterations = 20;   (* number of iterates of EM for each branch *)
      nmlngth = 10;      (* number of characters max. in species name *)
      epsilon = 0.00001; (* used in update *)

   TYPE base = (A, C, G, T);
      sitelike = ARRAY[A..T] OF REAL;
      phenotype = ARRAY [1..maxsites] OF sitelike;
      naym = ARRAY[1..nmlngth] OF CHAR;
      nodeptr = ^node;
      node = RECORD
         next, back : nodeptr;
         tip        : BOOLEAN;
         number     : INTEGER;
         x          : phenotype;
         nayme      : naym;
         v          : REAL;
         xcoord,
          ycoord    : INTEGER
         END;
      tree = RECORD
         nodep      : ARRAY[1..maxsp2] OF nodeptr;
         likelihood : REAL;
         start      : nodeptr
         END;

   VAR infile : TEXT;
      numsp, numsp1, numsp2, sites, outgrno : INTEGER;
      brief, outgropt, usertree, didreroot : BOOLEAN;
      curtree, bestree : tree;
      ttratio, freqa, freqc, freqg, freqt, freqr, freqy, freqar,
                                   freqcy, freqgr, freqty  : REAL;

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT,' Nucleic Acid Sequence Maximum Likelihood');
         WRITELN(OUTPUT, ' method, version 2.61');
         WRITELN(OUTPUT);
         READ(infile, numsp, sites);
         WRITELN(OUTPUT, ' ', numsp:4, ' Species, ', sites:4,
           ' Sites');
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      FUNCTION letter (ch : CHAR) : BOOLEAN;
      BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
         letter := ((ch >= 'A') AND (ch <= 'I'))
                   OR ((ch >= 'J') AND (ch <= 'R'))
                   OR ((ch >= 'S') AND (ch <= 'Z'))
                   OR ((ch >= 'a') AND (ch <= 'i'))
                   OR ((ch >= 'j') AND (ch <= 'r'))
                   OR ((ch >= 's') AND (ch <= 'z'));
      END; (* letter *)

      PROCEDURE uppercase (VAR ch : CHAR);
      BEGIN (* uppercase *)
         IF letter(ch) AND (ch >= 'a')
         THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
      END; (* uppercase *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
            i, optnum : INTEGER;
      BEGIN (* getoptions *)
         brief := FALSE;
         usertree := FALSE;
         optnum := 0;
         ttratio := 1.5;
         outgropt := FALSE;
         outgrno := 1;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'B') OR (ch = 'O') OR (ch = 'T') OR (ch = 'U')
            THEN CASE ch OF
                  'B' : brief := TRUE;
                  'O' : BEGIN
                        outgropt := TRUE;
                        optnum := optnum + 1;
                        END;
                  'T' : optnum := optnum + 1;
                  'U' : usertree := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         FOR i := 1 TO optnum DO BEGIN
            READLN(infile);
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'T'
            THEN READ(infile, ttratio);
            IF ch = 'O'
            THEN READ(infile, outgrno);
            END;
         ttratio := 1.0 / ttratio;
      END; (* getoptions *)

      PROCEDURE getbasefreqs;
      BEGIN (* getbasefreqs *)
         READLN(infile);
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' Base Frequencies:');
         WRITELN(OUTPUT, ' ');
         READ(infile, freqa);
         WRITELN(OUTPUT, '    A    ', freqa:10:5);
         READ(infile, freqc);
         WRITELN(OUTPUT, '    C    ', freqc:10:5);
         READ(infile, freqg);
         WRITELN(OUTPUT, '    G    ', freqg:10:5);
         READ(infile, freqt);
         WRITELN(OUTPUT, '   T(U)  ', freqt:10:5);
         freqr := freqa + freqg;
         freqy := freqc + freqt;
         freqar := freqa/freqr;
         freqcy := freqc/freqy;
         freqgr := freqg/freqr;
         freqty := freqt/freqy;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, '  Transition/transversion parameter = ',
                           1.0/ttratio:10:6);
         WRITELN(OUTPUT, '  Actual transition/transversion ratio = ',
            (freqa*freqgr + freqc*freqty + ttratio*(freqa*freqg
             + freqc*freqt))/(ttratio*freqy*freqr):10:6);
      END; (* getbasefreqs *)

      PROCEDURE setuptree (VAR a : tree);
         VAR i, j : INTEGER;
            p, q : nodeptr;
      BEGIN (* setuptree *)
         FOR i := 1 TO numsp DO BEGIN
            NEW(a.nodep[i]);
            a.nodep[i]^.tip := TRUE;
            a.nodep[i]^.number := i;
            END;
         FOR i := numsp1 TO numsp2 DO BEGIN
            q := NIL;
            FOR j := 1 TO 3 DO BEGIN
               NEW(p);
               p^.tip := FALSE;
               p^.number := i;
               p^.next := q;
               q := p;
               END;
            p^.next^.next^.next := p;
            a.nodep[i] := p;
            END;
         a.likelihood := -999999.0;
         a.start := a.nodep[1];
      END; (* setuptree *)

      PROCEDURE getdata;
         VAR i, j, k : INTEGER;
            ch : CHAR;
            b : base;
      BEGIN (* getdata *)
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' Name                       Sequences');
         WRITELN(OUTPUT, ' ----                       ---------');
         WRITELN(OUTPUT, ' ');
         setuptree (curtree);
         FOR i := 1 TO numsp DO BEGIN
            WRITE(OUTPUT, ' ');
            READLN(infile);
            FOR j := 1 TO nmlngth DO BEGIN
               READ(infile, ch);
               curtree.nodep[i]^.nayme[j] := ch;
               WRITE(OUTPUT, ch);
               END;
            WRITE(OUTPUT, '   ');
            FOR j := 1 TO sites DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
               uppercase (ch);
               IF NOT ((ch = 'A') OR (ch = 'C') OR (ch = 'G')
                     OR (ch = 'U') OR (ch = 'T') OR (ch = 'R')
                     OR (ch = 'Y') OR (ch = 'H') OR (ch = 'N')
                     OR (ch = 'X') OR (ch = '?') OR (ch = 'O')
                     OR (ch = ' '))
               THEN WRITELN(OUTPUT, ' Warning -- bad base: ', ch);
               WRITE(OUTPUT, ch);
               IF (j MOD 60) = 0
               THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR k := 1 TO nmlngth+4 DO
                     WRITE(OUTPUT, ' ');
                  END
               ELSE IF (j MOD 10) = 0 THEN WRITE(OUTPUT, ' ');
               FOR b := A TO T DO
                  curtree.nodep[i]^.x[j][b] := 0.0;
               CASE ch OF
                  'A' : curtree.nodep[i]^.x[j][A] := 1.0;
                  'C' : curtree.nodep[i]^.x[j][C] := 1.0;
                  'G' : curtree.nodep[i]^.x[j][G] := 1.0;
                  'T' : curtree.nodep[i]^.x[j][T] := 1.0;
                  'U' : curtree.nodep[i]^.x[j][T] := 1.0;
                  'R' : BEGIN
                        curtree.nodep[i]^.x[j][A] := 1.0;
                        curtree.nodep[i]^.x[j][G] := 1.0;
                        END;
                  'Y' : BEGIN
                        curtree.nodep[i]^.x[j][C] := 1.0;
                        curtree.nodep[i]^.x[j][T] := 1.0;
                        END;
                  'N' : FOR b := A TO T DO
                           curtree.nodep[i]^.x[j][b] := 1.0;
                  'X' : FOR b := A TO T DO
                           curtree.nodep[i]^.x[j][b] := 1.0;
                  '?' : FOR b := A TO T DO
                           curtree.nodep[i]^.x[j][b] := 1.0;
                  'H' : BEGIN
                        curtree.nodep[i]^.x[j][A] := 1.0;
                        curtree.nodep[i]^.x[j][C] := 1.0;
                        curtree.nodep[i]^.x[j][T] := 1.0;
                        END
                  END;
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
         IF (NOT usertree)
         THEN setuptree(bestree);
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      getoptions;
      getbasefreqs;
      getdata;
   END; (* getinput *)

   PROCEDURE maketree;
      VAR nextsp, numtrees, which : INTEGER;
         succeeded, smoothed : BOOLEAN;

      FUNCTION evaluate( VAR tr : tree ) : REAL;
         VAR sum, y, z, y1, z1, prod12, prod1, prod2, prod3 : REAL;
            i : INTEGER;
            p, q : nodeptr;
            x1, x2 : sitelike;
      BEGIN (* evaluate *)
         sum := 0.0;
         p := tr.start;
         q := p^.back;
         y := p^.v;
         z := 1.0 - y;
         IF z > 0.0
         THEN z1 := EXP(ttratio*LN(z))
         ELSE z1 := 0.0;
         y1 := 1.0 - z1;
         FOR i := 1 TO sites DO BEGIN
            x1 := p^.x[i];
            prod1 := freqa*x1[A] + freqc*x1[C]
                      + freqg*x1[G] + freqt*x1[T];
            x2 := q^.x[i];
            prod2 := freqa*x2[A] + freqc*x2[C]
                      + freqg*x2[G] + freqt*x2[T];
            prod3 := (x1[A]*freqa+x1[G]*freqg)*(x2[A]*freqar+x2[G]*freqgr)
                         + (x1[C]*freqc+x1[T]*freqt)
                                 *(x2[C]*freqcy+x2[T]*freqty);
            prod12 := freqa*x1[A]*x2[A] + freqc*x1[C]*x2[C]
                      + freqg*x1[G]*x2[G] + freqt*x1[T]*x2[T];
            sum := sum + LN(z1*(z*prod12 + y*prod3)+y1*prod1*prod2);
            END;
         tr.likelihood := sum;
         evaluate := sum;
      END; (* evaluate *)

      PROCEDURE nuview (p : nodeptr);
         VAR i : INTEGER;
            v1, v2, w1, w2, z1, z2, y1, y2, sum1, sum2, sumr1, sumr2,
                                             sumy1, sumy2 : REAL;
            q, r : nodeptr;
            xx1, xx2, xx3 : sitelike;
      BEGIN (* nuview *)
         q := p^.next^.back;
         r := p^.next^.next^.back;
         v1 := q^.v;
         v2 := r^.v;
         w1 := 1.0 - v1;
         w2 := 1.0 - v2;
         IF w1 <= 0.0
         THEN z1 := 0.0
         ELSE z1 := EXP(ttratio*LN(w1));
         IF w2 <= 0.0
         THEN z2 := 0.0
         ELSE z2 := EXP(ttratio*LN(w2));
         y1 := 1.0 - z1;
         y2 := 1.0 - z2;
         FOR i := 1 TO sites DO BEGIN
            xx1 := q^.x[i];
            xx2 := r^.x[i];
            sum1 := y1*(freqa*xx1[A] + freqc*xx1[C]
                     + freqg*xx1[G] + freqt*xx1[T]);
            sum2 := y2*(freqa*xx2[A] + freqc*xx2[C]
                     + freqg*xx2[G] + freqt*xx2[T]);
            sumr1 :=v1*(freqar*xx1[A] + freqgr*xx1[G]);
            sumr2 :=v2*(freqar*xx2[A] + freqgr*xx2[G]);
            sumy1 :=v1*(freqcy*xx1[C] + freqty*xx1[T]);
            sumy2 :=v2*(freqcy*xx2[C] + freqty*xx2[T]);
            xx3[A] := (sum1 + z1*(w1*xx1[A] + sumr1))
                        *(sum2 + z2*(w2*xx2[A] + sumr2));
            xx3[C] := (sum1 + z1*(w1*xx1[C] + sumy1))
                        *(sum2 + z2*(w2*xx2[C] + sumy2));
            xx3[G] := (sum1 + z1*(w1*xx1[G] + sumr1))
                        *(sum2 + z2*(w2*xx2[G] + sumr2));
            xx3[T] := (sum1 + z1*(w1*xx1[T] + sumy1))
                        *(sum2 + z2*(w2*xx2[T] + sumy2));
            p^.x[i] := xx3;
            END;
      END; (* nuview *)

      PROCEDURE update (p : nodeptr);

         PROCEDURE makenewv (p : nodeptr);
            VAR i, it : INTEGER;
               sum, sum1, sum2, y, z, y1, z1, yprod, yold : REAL;
               prod, prod2, prod3 : ARRAY[1..maxsites] OF REAL;
               done : BOOLEAN;
               q : nodeptr;
               xx1, xx2 : sitelike;
         BEGIN (* makenewv *)
            q := p^.back;
            y := p^.v;
            FOR i := 1 TO sites DO BEGIN
               xx1 := p^.x[i];
               xx2 := q^.x[i];
               sum1 := freqa*xx1[A] + freqc*xx1[C]
                         + freqg*xx1[G] + freqt*xx1[T];
               sum2 := freqa*xx2[A] + freqc*xx2[C]
                         + freqg*xx2[G] + freqt*xx2[T];
               prod[i] := sum1*sum2;
               prod2[i] := (xx1[A]*freqa+xx1[G]*freqg)*(xx2[A]*freqar
                            +xx2[G]*freqgr) + (xx1[C]*freqc+xx1[T]*freqt)
                            *(xx2[C]*freqcy+xx2[T]*freqty);
               prod3[i] := freqa*xx1[A]*xx2[A] + freqc*xx1[C]*xx2[C]
                            + freqg*xx1[G]*xx2[G] + freqt*xx1[T]*xx2[T];
               END;
            IF y < epsilon
            THEN y := 10.0*epsilon;
            IF y = 1.0
            THEN y := 1.0 - 10.0*epsilon;
            done := FALSE;
            it := 1;
            WHILE (it < iterations) AND (NOT done) DO BEGIN
               z := 1.0 - y;
               IF z > 0.0
               THEN z1 := EXP(ttratio*LN(z))
               ELSE z1 := 0.0;
               y1 := 1.0 - z1;
               sum := 0.0;
               FOR i := 1 TO sites DO BEGIN
                  yprod := y*z1*prod2[i];
                  sum := sum + (yprod+y*ttratio*prod[i])
                                  /(z1*z*prod3[i]+yprod+y1*prod[i])
                  END;
               yold := y;
               y := sum/((1.0+ttratio)*sites);
               done := ABS(yold-y) < epsilon;
               it := it + 1;
               END;
            smoothed := smoothed AND done;
            p^.v := y;
            q^.v := y;
         END; (* makenewv *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN nuview(p);
         IF NOT p^.back^.tip
         THEN nuview(p^.back);
         makenewv(p);
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
      BEGIN (* smooth *)
         update (p);
         IF NOT p^.tip
         THEN BEGIN
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE hookup (p, q : nodeptr);
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      PROCEDURE insert (p, q: nodeptr);
         VAR i : INTEGER;
            r : nodeptr;
      BEGIN (* insert *)
         r := p^.next^.next;
         hookup (r, q^.back);
         hookup (p^.next, q);
         q^.v := 0.5 * q^.v;
         q^.back^.v := q^.v;
         r ^.v := q^.v;
         r^.back^.v := r^.v;
         smoothed := FALSE;
         i := 1;
         WHILE (i < smoothings) AND (NOT smoothed) DO BEGIN
            smoothed := TRUE;
            smooth (p);
            smooth (p^.back);
            i := i + 1;
            END;
      END; (* insert *)

      PROCEDURE remove ( VAR p, q : nodeptr);
      BEGIN (* remove *)
         q := p^.next^.back;
         hookup (q, p^.next^.next^.back);
         p^.next^.back := NIL;
         p^.next^.next^.back := NIL;
         update (q);
         update (q^.back);
      END; (* remove *)

      PROCEDURE copy ( VAR a, b : tree);
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO
            b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE buildnewtip (m : INTEGER; VAR tr : tree);
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         p := tr.nodep[m + numsp - 2];
         hookup (tr.nodep[m], p);
         p^.v := 0.5;
         p^.back^.v := 0.5;
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR tr : tree);
      BEGIN (* buildsimpletree *)
         hookup (tr.nodep[1], tr.nodep[2]);
         tr.nodep[1]^.v := 0.5;
         tr.nodep[2]^.v := 0.5;
         buildnewtip (3, tr);
         insert(tr.nodep[3]^.back, tr.nodep[1]);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr);
      BEGIN (* addtraverse *)
         insert(p, q);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN copy (curtree, bestree);
         remove(p, q);
         IF NOT q^.tip
         THEN BEGIN
            addtraverse (p, q^.next^.back);
            addtraverse (p, q^.next^.next^.back);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p, q : nodeptr);

         PROCEDURE tryrearr (p, q : nodeptr);
            VAR r : nodeptr;
         BEGIN (* tryrearr *)
            r := q;
            remove (p, r);
            insert (p, q);
            numtrees := numtrees + 1;
            IF evaluate(curtree) > bestree.likelihood
            THEN BEGIN
               succeeded := TRUE;
               copy (curtree, bestree);
               END
            ELSE copy (bestree, curtree);
         END; (* tryrearr *)

      BEGIN (* rearrange *)
         IF NOT succeeded
         THEN BEGIN
            IF (NOT p^.tip) AND (NOT q^.tip)
            THEN BEGIN
               tryrearr (p, q^.next);
               tryrearr (p, q^.next^.next);
               END;
            IF NOT q^.tip
            THEN BEGIN
               rearrange (q, q^.next^.back);
               rearrange (q, q^.next^.next^.back);
               END;
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + over;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.next^.back);
               p^.xcoord := p^.next^.back^.xcoord;
               coordinates (p^.next^.next^.back);
               p^.ycoord := p^.next^.next^.back^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            WRITE(OUTPUT, ' ');
            REPEAT
               IF p^.next^.back^.ycoord >= i
               THEN p := p^.next^.back
               ELSE BEGIN
                  q := p^.next^.next^.back;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.number > 9
                     THEN WRITE(OUTPUT, p^.number:2)
                     ELSE WRITE(OUTPUT, p^.number:1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO WRITE(OUTPUT, p^.nayme[j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         coordinates (curtree.start^.back);
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO tipy-down DO drawline (i);
         FOR i := 1 TO down DO WRITELN(OUTPUT, ' !');
         WRITE(OUTPUT, ' ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, curtree.start^.nayme[i]);
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE summarize;

         PROCEDURE describe (p : nodeptr);
            VAR i : INTEGER;
               q : nodeptr;
               s, sumlr : REAL;

            FUNCTION sigma(p : nodeptr) : REAL;
               VAR sum, y, suma, sumb, sumc1, sumc2, r, d2, d,
                               li, yr, yrm1, yrm2, ab, bc : REAL;
                  i : INTEGER;
                  q : nodeptr;
                  xx1, xx2 : sitelike;
            BEGIN (* sigma *)
               sum := 0.0;
               sumlr := 0.0;
               q := p^.back;
               y := 1.0 - p^.v;
               r := ttratio;
               yrm2 := EXP((r-2.0)*LN(y));
               yrm1 := yrm2*y;
               yr := yrm1*y;
               FOR i := 1 TO sites DO BEGIN
                  xx1 := p^.x[i];
                  xx2 := q^.x[i];
                  sumc1 := freqa*xx1[A] + freqc*xx1[C]
                            + freqg*xx1[G] + freqt*xx1[T];
                  sumc2 := freqa*xx2[A] + freqc*xx2[C]
                            + freqg*xx2[G] + freqt*xx2[T];
                  sumb := (xx1[A]*freqa+xx1[G]*freqg)*(xx2[A]*freqar
                           +xx2[G]*freqgr) + (xx1[C]*freqc+xx1[T]*freqt)
                            *(xx2[C]*freqcy+xx2[T]*freqty);
                  suma := freqa*xx1[A]*xx2[A] + freqc*xx1[C]*xx2[C]
                            + freqg*xx1[G]*xx2[G] + freqt*xx1[T]*xx2[T];
                  ab := suma - sumb;
                  bc := sumb - sumc1*sumc2;
                  li := sumc1*sumc2 + yr*(y*ab + bc);
                  d := (r+1.0)*yr*ab + r*yrm1*bc;
                  d2 := r*((r+1.0)*yrm1*ab + (r-1.0)*yrm2*bc);
                  sum := sum + (SQR(d/li) - d2/li);
                  IF suma > 0.0
                  THEN sumlr := sumlr + LN(li/suma)
                  ELSE sumlr := 1.0E+10;
                  END;
               IF sum > 1.0E-6
               THEN sigma := 1.95996 * SQRT(1.0/sum)
               ELSE sigma := -1.0;
            END; (* sigma *)

         BEGIN (* describe *)
            q := p^.back;
            WRITE(OUTPUT, ' ');
            IF q^.tip
            THEN FOR i := 1 TO nmlngth DO WRITE(OUTPUT, q^.nayme[i])
            ELSE WRITE(OUTPUT, q^.number:4, '      ');
            WRITE(OUTPUT, '    ');
            IF p^.tip
            THEN FOR i := 1 TO nmlngth DO WRITE(OUTPUT, p^.nayme[i])
            ELSE WRITE(OUTPUT, p^.number:4, '      ');
            WRITE(OUTPUT, q^.v:15:5);
            s := sigma(q);
            IF s >= 0.0
            THEN WRITE(OUTPUT, '     (', q^.v-s:9:5, ',',
                                            q^.v+s:12:5, ')')
            ELSE WRITE(OUTPUT, '     ( -infinity,   infinity)');
            IF sumlr > 1.9205
            THEN WRITE(OUTPUT, ' *');
            IF sumlr > 2.995
            THEN WRITE(OUTPUT, '*');
            WRITELN(OUTPUT);
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, ' remember: ');
         IF didreroot
         THEN WRITE(OUTPUT, '(although rooted by outgroup) ');
         WRITELN(OUTPUT, 'this is an unrooted tree!');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Ln Likelihood = ', curtree.likelihood:11:5);
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT, ' ');
            WRITELN(OUTPUT, ' Examined ',numtrees:4, ' trees');
            END;
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, '  Between        And            Length');
         WRITELN(OUTPUT, '      Approx. Confidence Limits');
         WRITE(OUTPUT, '  -------        ---            ------');
         WRITELN(OUTPUT, '      ------- ---------- ------');
         describe (curtree.start^.back);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, '      *  = significantly positive, P < 0.05');
         WRITELN(OUTPUT, '      ** = significantly positive, P < 0.01');
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' ');
      END; (* summarize *)

      PROCEDURE treeread;
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            VAR q : nodeptr;
               n : INTEGER;
         BEGIN (* addelement *)
            REPEAT
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               hookup(p, q);
               addelement (q^.next);
               findch (',');
               addelement (q^.next^.next);
               findch (')');
               READ(infile, ch);
               END
            ELSE BEGIN
               n := 0;
               REPEAT
                  n := 10*n + ORD(ch) - ORD('0');
                  READ(infile, ch);
               UNTIL NOT((ORD(ch)>=ORD('0')) AND (ORD(ch)<=ORD('9')));
               hookup (curtree.nodep[n], p);
               IF curtree.start^.number > n
               THEN curtree.start := curtree.nodep[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               addelement (p);
               p := p^.next;
               findch (',');
               END;
            addelement (p);
            FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
            END;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE treevaluate;
         VAR dummy : REAL;

         PROCEDURE nodeinit (p : nodeptr);
            VAR q, r : nodeptr;
               i : INTEGER;
               b : base;
         BEGIN (* nodeinit *)
            IF NOT p^.tip
            THEN BEGIN
               q := p^.next^.back;
               r := p^.next^.next^.back;
               nodeinit (q);
               nodeinit (r);
               FOR i := 1 TO sites DO
                  FOR b := A TO T DO
                     p^.x[i][b] := 0.5 * (q^.x[i][b] + r^.x[i][b]);
               END;
         END; (* nodeinit *)

      BEGIN (* treevaluate *)
         nodeinit (curtree.start);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         READLN(infile, numtrees);
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            treevaluate;
            printree;
            summarize;
            END;
         END
      ELSE BEGIN
         buildsimpletree (curtree);
         numtrees := 1;
         nextsp := 4;
         WHILE nextsp <= numsp DO BEGIN
            buildnewtip (nextsp, curtree);
            bestree.likelihood := -999999.0;
            addtraverse (curtree.nodep[nextsp]^.back,
                                  curtree.start^.back);
            copy (bestree, curtree);
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start, curtree.start^.back);
               END;
            IF (NOT brief) OR (nextsp = numsp)
            THEN BEGIN
               didreroot := outgropt AND (nextsp = numsp);
               IF didreroot
               THEN curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               END;
            nextsp := nextsp + 1;
            END;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   RESET(infile);
   getinput;
   maketree;
END.
