PROGRAM Kitsch (infile, OUTPUT);
  (* version 2.6 *)

CONST maxsp = 25;    (* maximum number of species               *)
   maxsz = 49;       (* size of pointer array.  >= 2*maxsp - 1  *)
   namelength = 10;  (* number of characters in species name    *)

TYPE vector = ARRAY [1..maxsz] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;
   pointer = ^node;           (* nodes will form binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : pointer;
      index : INTEGER;
      tip : BOOLEAN;        (* present species are tips of tree *)
      d, w : vector;        (* distances and weights *)
      t : REAL;             (* time *)
      sametime : BOOLEAN;   (* bookkeeps scrunched nodes *)
      weight : REAL;        (* weight of node used by scrunch *)
      processed : BOOLEAN;  (* used by evaluate *)
      xcoord, ycoord : INTEGER  (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF pointer;

VAR root : pointer;
   infile : TEXT;
   numsp, numsp2, numtrees, power : INTEGER;
     (* numsp = number of species
        numtrees is used by usertree option part of maketree *)
   treenode : pointarray;      (* pointers to all nodes in tree *)
   names : ARRAY[1..maxsp] OF naym;
   brief, usertree, lower, upper, negallowed, powers : BOOLEAN;

   FUNCTION letter (ch : CHAR) : BOOLEAN;
   BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
   BEGIN (* uppercase *)
      IF letter(ch) AND (ch >= 'a')
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT,' Fitch-Margoliash method ');
         WRITE(OUTPUT, 'with contemporary tips,');
         WRITELN(OUTPUT, ' version 2.6');
         WRITELN(OUTPUT, ' ');
         READ(infile, numsp);
         WRITELN(OUTPUT, ' ', numsp:4, ' Populations');
         numsp2 := 2*numsp - 1;
      END; (* getnums *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
      BEGIN (* getoptions *)
         brief := FALSE;
         usertree := FALSE;
         lower := FALSE;
         upper := FALSE;
         negallowed := FALSE;
         powers := FALSE;
         power := 2;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'B') OR (ch = 'L') OR (ch = 'N') OR (ch = 'P')
               OR (ch = 'R') OR (ch = 'U')
            THEN CASE ch OF
                  'B' : brief := TRUE;
                  'L' : lower := TRUE;
                  'N' : negallowed := TRUE;
                  'P' : powers := TRUE;
                  'R' : upper := TRUE;
                  'U' : usertree := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         IF powers
         THEN BEGIN
            READLN(infile);
            READ(infile, ch);
            READ(infile, power);
            END;
         WRITELN(OUTPUT, '                                     2');
         WRITELN(OUTPUT, '                   <  <   (Obs - Exp)');
         WRITELN(OUTPUT, ' Sum of squares =  <  <   ------------');
         WRITELN(OUTPUT, '                   i  j         ',
                    power:2);
         WRITELN(OUTPUT, '                              Obs');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' negative branch lengths');
         IF NOT negallowed THEN WRITE(OUTPUT, ' not');
         WRITELN(OUTPUT, ' allowed');
      END; (* getoptions *)

      PROCEDURE getdata;
         VAR i, j, k : INTEGER;
            skipit, skipother : BOOLEAN;
            x : REAL;
      BEGIN (* getdata *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Name                       Distances');
         WRITELN(OUTPUT, ' ----                       ---------');
         WRITELN(OUTPUT);
         FOR i := 1 TO numsp2 DO BEGIN
            NEW(treenode[i]);
            treenode[i]^.ancestor := NIL;
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.index := i;
            treenode[i]^.tip := (i <= numsp);
            treenode[i]^.t := 0.0;
            treenode[i]^.sametime := FALSE;
            END;
         FOR i := 1 TO numsp DO BEGIN
            treenode[i]^.d[i] := 0.0;
            treenode[i]^.w[i] := 0.0;
            treenode[i]^.weight := 0.0;
            READLN(infile);
            FOR j := 1 TO namelength DO READ(infile, names[i][j]);
            FOR j := 1 TO numsp DO BEGIN
               skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
               skipother := (lower AND (i >= j)) OR
                              (upper AND (i <= j));
               IF NOT skipit
               THEN BEGIN
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, x);
                  treenode[i]^.d[j] := x;
                  IF skipother
                  THEN treenode[j]^.d[i] := treenode[i]^.d[j];
                  END;
               END;
            END;
         FOR i := 1 TO numsp DO BEGIN
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO namelength DO WRITE(OUTPUT, names[i][j]);
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO numsp DO BEGIN
               WRITE(OUTPUT, treenode[i]^.d[j]:10:5);
               IF (j MOD 6) = 0 THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR k := 1 TO namelength+2 DO
                     WRITE(OUTPUT, ' ');
                  END;
               END;
            WRITELN(OUTPUT);
            END;
         FOR i := 1 TO numsp DO
            FOR j := 1 TO numsp DO
               IF i <> j
               THEN BEGIN
                  x := 1.0;
                  FOR k := 1 TO power DO
                     x := x / treenode[i]^.d[j];
                  treenode[i]^.w[j] := x;
                  END;
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      getoptions;
      getdata;
   END; (* getinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      VAR i, examined : INTEGER;
         like, bestyet : REAL;
         there, item : pointer;

      PROCEDURE add (VAR below, newtip, newfork : pointer);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : pointer);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : pointer;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: pointer);
         (* fill in times and evaluate sum of squares for tree *)
         VAR i : INTEGER;
            sum : REAL;

         PROCEDURE firstraverse (q : pointer);
            (* prune the tree *)
            VAR i, j, k : INTEGER;
               u, v : pointer;

            PROCEDURE scrunch (s : pointer);
               (* see if nodes can be combined to
                  prevent negative lengths *)
               VAR tmax : REAL;
                  closest : pointer;
                  found : BOOLEAN;

               PROCEDURE scrunchtraverse (u : pointer);
                  (* traverse to find closest node to
                     the current one *)
               BEGIN (* scrunchtraverse *)
                  IF NOT u^.sametime
                  THEN BEGIN
                     IF u^.t > tmax
                     THEN BEGIN
                        closest := u;
                        tmax := u^.t;
                        END;
                     END
                  ELSE BEGIN
                     u^.t := u^.ancestor^.t;
                     IF NOT u^.tip
                     THEN BEGIN
                        scrunchtraverse (u^.leftdesc);
                        scrunchtraverse (u^.rtdesc);
                        END;
                     END;
               END; (* scrunchtraverse *)

               PROCEDURE combine (a, b : pointer);
                  (* put node b into the set having the same
                     time as a *)
               BEGIN (* combine *)
                  a^.t := (a^.t * a^.weight + b^.t * b^.weight)
                           / (a^.weight + b^.weight);
                  a^.weight := a^.weight + b^.weight;
                  b^.sametime := TRUE;
               END; (* combine *)

            BEGIN (* scrunch *)
               closest := NIL;
               tmax := -1.0;
               REPEAT
                  scrunchtraverse (s^.leftdesc);
                  scrunchtraverse (s^.rtdesc);
                  found := (tmax > s^.t);
                  IF found THEN combine (s, closest);
                  tmax := -1.0;
               UNTIL (NOT found);
            END; (* scrunch *)

            PROCEDURE secondtraverse (a : pointer);
               (* recalculate distances, add to sum *)
               VAR l : INTEGER;
                  wil, wjl, wkl, wli, wlj, wlk : REAL;
            BEGIN (* secondtraverse *)
               IF NOT a^.processed
               THEN BEGIN
                  secondtraverse (a^.leftdesc);
                  secondtraverse (a^.rtdesc);
                  END
               ELSE IF a <> q
                  THEN BEGIN
                     l := a^.index;
                     wil := u^.w[l];
                     wjl := v^.w[l];
                     wkl := wil + wjl;
                     wli := a^.w[i];
                     wlj := a^.w[j];
                     wlk := wli + wlj;
                     q^.w[l] := wkl;
                     a^.w[k] := wlk;
                     q^.d[l] := (wil * a^.d[i] + wjl * a^.d[j]) / wkl;
                     a^.d[k] := (wli * u^.d[l] + wlj * v^.d[l]) / wlk;
                     sum := sum + (wil*wjl/wkl)*SQR(a^.d[i]-a^.d[j])
                                + (wli*wlj/wlk)*SQR(u^.d[l]-v^.d[l]);
                     END;
            END; (* secondtraverse *)

         BEGIN (* firsttraverse *)
            IF q <> NIL
            THEN BEGIN
               q^.sametime := FALSE;
               firstraverse (q^.leftdesc);
               firstraverse (q^.rtdesc);
               q^.processed := TRUE;
               IF NOT q^.tip
               THEN BEGIN
                  u := q^.leftdesc;
                  v := q^.rtdesc;
                  i := u^.index;
                  j := v^.index;
                  k := q^.index;
                  q^.t :=  u^.d[j]/2.0;
                  q^.weight := u^.w[j];
                  IF NOT negallowed THEN scrunch (q);
                  secondtraverse (r);
                  END;
               END;
         END; (* firstraverse *)

         PROCEDURE sumtraverse (q : pointer);
            (* traverse to finish computation of sum
               of squares *)
            VAR i, j : INTEGER;
               u, v : pointer;
         BEGIN (* sumtraverse *)
            IF NOT q^.tip
            THEN BEGIN
               sumtraverse (q^.leftdesc);
               sumtraverse (q^.rtdesc);
               u := q^.leftdesc;
               v := q^.rtdesc;
               i := u^.index;
               j := v^.index;
               sum := sum + u^.w[j]*SQR(u^.d[j]-2.0*q^.t)
                          + v^.w[i]*SQR(v^.d[i]-2.0*q^.t);
               END;
         END; (* sumtraverse *)

      BEGIN (* evaluate *)
         sum := 0.0;
         FOR i := 1 TO numsp2 DO
            treenode[i]^.processed := treenode[i]^.tip;
         firstraverse (r);
         sumtraverse (r);
         examined := examined + 1;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE addpreorder (p, item : pointer);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)

         PROCEDURE tryadd (p: pointer);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR nufork : pointer;
         BEGIN (* tryadd *)
            nufork := treenode[numsp+i-1];
            add (p, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            evaluate (root);
            remove (item, nufork);
            IF nufork = root THEN root := p;
            IF like > bestyet THEN BEGIN
               bestyet := like;
               there := p;
               END;
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item);
            addpreorder (p^.rtdesc, item);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: pointer);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", procedure rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : pointer);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : pointer;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := like;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     like := oldlike;
                     END
                  ELSE BEGIN
                     bestyet := like;
                     success := TRUE;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : pointer);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (VAR p : pointer);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : pointer;
               n : INTEGER;
         BEGIN (* addelement *)
            REPEAT
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := treenode[nextnode];
               addelement (q^.leftdesc);
               q^.leftdesc^.ancestor := q;
               findch (',');
               IF EOLN(infile) THEN READLN(infile);
               addelement (q^.rtdesc);
               q^.rtdesc^.ancestor := q;
               findch (')');
               p := q;
               IF NOT EOLN(infile) THEN READ(infile, ch);
               END
            ELSE BEGIN
               n := 0;
               REPEAT
                  n := 10*n + ORD(ch) - ORD('0');
                  READ(infile, ch);
               UNTIL NOT ((ch >= '0') AND (ch <= '9'));
               p := treenode[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[numsp + 1];
         nextnode := numsp;
         root^.ancestor := NIL;
         addelement (root);
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : pointer);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : pointer;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p :=root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.index >= 10
                     THEN WRITE(OUTPUT, p^.index:2)
                     ELSE WRITE(OUTPUT, p^.index:1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO namelength DO
                  WRITE(OUTPUT, names[p^.index][j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV numsp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITELN(OUTPUT, '  /');
         WRITELN(OUTPUT, ' /');
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints number of steps and table of numbers of steps in
            each character *)

         PROCEDURE dtraverse (q : pointer);
            (* print table of lengths etc. *)
            VAR i : INTEGER;
         BEGIN (* dtraverse *)
            IF NOT q^.tip
            THEN BEGIN
               dtraverse(q^.leftdesc);
               dtraverse(q^.rtdesc);
               END;
            IF q^.ancestor <> NIL
            THEN BEGIN
               WRITE(OUTPUT, q^.ancestor^.index:4, '  ');
               IF q^.index <= numsp
               THEN FOR i := 1 TO namelength DO
                  WRITE(OUTPUT, names[q^.index][i])
               ELSE WRITE(OUTPUT, q^.index:4, '      ');
               WRITE(OUTPUT, (q^.ancestor^.t-q^.t):13:5);
               WRITELN(OUTPUT, (root^.t-q^.t):15:5);
               END;
         END; (* dtraverse *)

      BEGIN (* describe *)
         WRITELN(OUTPUT, ' Sum of squares = ',-like:10:3);
         WRITELN(OUTPUT);
         IF power = 2
         THEN BEGIN
            WRITE(OUTPUT, ' Average percent standard deviation =');
            WRITELN(OUTPUT,
               100*SQRT(-like/((numsp*(numsp-1))-2)):10:5);
            WRITELN(OUTPUT);
            END;
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT, ' examined ',examined:4, ' trees');
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT, ' from    to           length          time');
         WRITELN(OUTPUT, ' ----    --           ------          ----');
         WRITELN(OUTPUT);
         dtraverse (root);
         WRITELN(OUTPUT);
      END; (* describe *)

   BEGIN (* maketree *)
      IF NOT usertree
      THEN BEGIN
         examined := 0;
         add (treenode[2], treenode[1], treenode[numsp+1]);
         root := treenode[numsp+1];
         FOR i:= 3 TO numsp DO BEGIN
            bestyet := -10000000.0;
            there := root;
            item := treenode[i];
            addpreorder (root, item);
            add (there, treenode[i], treenode[numsp+i-1]);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            like := bestyet;
            rearrange (root);
            evaluate (root);
            examined := examined - 1;
            IF (NOT brief) OR (i = numsp)
            THEN BEGIN
               printree;
               describe;
               END;
            END;
         END
      ELSE BEGIN
         READLN(infile, numtrees);
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' User-defined trees:');
         FOR i := 1 TO numtrees DO BEGIN
            treeread;
            evaluate (root);
            printree;
            describe;
            END;
         END;
   END; (* maketree *)

BEGIN (* Fitch-Margoliash criterion with contemporary tips *)
   (* reads in numsp, options, and the data, then calls maketree to
      construct the tree *)
   RESET(infile);
   getinput;
   maketree;
  (*  close(infile);  -- may need this here *)
END. (* Fitch-Margoliash criterion with contemporary tips *)
